@@ -19,16 +19,22 @@
  ***************************************************************************/
 
 #include <QApplication>
+#include <QProcessEnvironment>
 #include <QDebug>
 #include <QEventLoop>
 #include <QTimer>
 #include <QDir>
+#include <unistd.h>
 
 #include "mprocess.h"
 
 MProcess::MProcess(QObject *parent)
     : QProcess(parent)
 {
+    // Stop user-selected locale from interfering with command output parsing.
+    QProcessEnvironment env = QProcessEnvironment::systemEnvironment();
+    env.insert("LC_ALL", "C.UTF-8");
+    setProcessEnvironment(env);
 }
 
 void MProcess::setupUI(QListWidget *listLog, QProgressBar *progInstall)
@@ -41,20 +47,26 @@ void MProcess::setupUI(QListWidget *listLog, QProgressBar *progInstall)
     listLog->setPalette(pal);
 }
 
-bool MProcess::exec(const QString &cmd, const bool rawexec, const QByteArray *input, bool needRead)
+void MProcess::setupChildProcess()
+{
+    if (chRoot.isEmpty()) return;
+    chroot(chRoot.toUtf8().constData());
+    chdir("/");
+}
+void MProcess::setChRoot(const QString &newroot)
+{
+    if (!newroot.isEmpty()) log("Set chroot: " + newroot, Standard);
+    else if (!chRoot.isEmpty() && newroot.isEmpty()) log("End chroot: " + chRoot, Standard);
+    chRoot = newroot;
+}
+
+bool MProcess::exec(const QString &program, const QStringList &arguments,
+    const QByteArray *input, bool needRead, QListWidgetItem *logEntry)
 {
-    if (halting) return false;
-    ++execount;
-    qDebug().nospace().noquote() << "Exec #" << execount << ": " << cmd;
-    QListWidgetItem *logEntry = log(cmd, Exec);
     QEventLoop eloop;
+    connect(this, QOverload<QProcess::ProcessError>::of(&QProcess::errorOccurred), &eloop, &QEventLoop::quit);
     connect(this, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), &eloop, &QEventLoop::quit);
-    if (!rawexec) start("/bin/bash", {"-c", cmd});
-    else {
-        QStringList args = splitCommand(cmd);
-        QString prog = args.takeFirst();
-        start(prog, args);
-    }
+    start(program, arguments);
     if (!debugUnusedOutput) {
         if (!needRead) closeReadChannel(QProcess::StandardOutput);
         closeReadChannel(QProcess::StandardError);
@@ -63,6 +75,8 @@ bool MProcess::exec(const QString &cmd, const bool rawexec, const QByteArray *in
     closeWriteChannel();
     eloop.exec();
     disconnect(this, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished), nullptr, nullptr);
+    disconnect(this, QOverload<QProcess::ProcessError>::of(&QProcess::errorOccurred), nullptr, nullptr);
+
     if (debugUnusedOutput) {
         bool hasOut = false;
         if (!needRead) {
@@ -83,25 +97,41 @@ bool MProcess::exec(const QString &cmd, const bool rawexec, const QByteArray *in
             logEntry->setFont(logFont);
         }
     }
-    qDebug().nospace() << "Exit #" << execount << ": " << exitCode() << " " << exitStatus();
+    qDebug().nospace() << "Exit #" << execount << ": " << exitCode() << " " << exitStatus() << " " << error();
+
     int status = 1;
     if (exitStatus() != QProcess::NormalExit) status = -1;
+    else if (error() != QProcess::UnknownError) status = -2;
     else if (exitCode() != 0) status = 0;
+
     log(logEntry, status);
     return (status > 0);
 }
 
-QString MProcess::execOut(const QString &cmd, bool everything)
+bool MProcess::exec(const QString &program, const QStringList &arguments, const QByteArray *input, bool needRead)
+{
+    if (halting) return false;
+    ++execount;
+    const QString &cmd = joinCommand(program, arguments);
+    qDebug().nospace().noquote() << "Exec #" << execount << ": " << cmd;
+    return exec(program, arguments, input, needRead, log(cmd, Exec));
+}
+bool MProcess::shell(const QString &cmd,  const QByteArray *input, bool needRead)
+{
+    if (halting) return false;
+    ++execount;
+    qDebug().nospace().noquote() << "Bash #" << execount << ": " << cmd;
+    return exec("/bin/bash", {"-c", cmd}, input, needRead, log(cmd, Exec));
+}
+
+QString MProcess::readOut(bool everything)
 {
-    exec(cmd, false, nullptr, true);
     QString strout(readAllStandardOutput().trimmed());
     if (everything) return strout;
     return strout.section("\n", 0, 0);
 }
-
-QStringList MProcess::execOutLines(const QString &cmd, const bool rawexec)
+QStringList MProcess::readOutLines()
 {
-    exec(cmd, rawexec, nullptr, true);
     return QString(readAllStandardOutput().trimmed()).split('\n', Qt::SkipEmptyParts);
 }
 
@@ -110,6 +140,7 @@ void MProcess::halt()
     halting = true;
     terminate();
     QTimer::singleShot(5000, this, &QProcess::kill);
+    setChRoot();
 }
 
 void MProcess::unhalt()
@@ -147,17 +178,18 @@ QString MProcess::joinCommand(const QString &program, const QStringList &argumen
 QListWidgetItem *MProcess::log(const QString &text, const LogType type)
 {
     if (type == Standard) qDebug().noquote() << text;
-    if (type == Section) qDebug().noquote() << "+++" << text << "+++";
-    else if (type == Status) qDebug().noquote() << "-" << text << "-";
-    if (!logView) return nullptr;
+    else if (type == Section) qDebug().noquote() << "<<" << text << ">>";
+    else if (type == Status) qDebug().noquote() << "++" << text << "++";
+    else if (type == Fail) qDebug().noquote() << "--" << text << "--";
+    if (!logView || type == Section) return nullptr;
     QListWidgetItem *entry = new QListWidgetItem(text, logView);
     logView->addItem(entry);
     if (type == Exec) entry->setForeground(Qt::cyan);
-    else if (type != Standard) {
+    else if (type == Status || type == Fail) {
         QFont font(entry->font());
-        if (type == Section) font.setBold(true);
-        else if (type == Status) font.setItalic(true);
+        font.setBold(true);
         entry->setFont(font);
+        if (type == Fail) entry->setForeground(Qt::magenta);
     }
     logView->scrollToBottom();
     return entry;
@@ -234,3 +266,41 @@ bool MProcess::mkpath(const QString &path)
     log(logEntry, rc ? 1 : -1);
     return rc;
 }
+
+const QString &MProcess::detectArch()
+{
+    if (testArch.isEmpty()) {
+        if (exec("uname", {"-m"}, nullptr, true)) testArch = readOut();
+        qDebug() << "Detect arch:" << testArch;
+    }
+    return testArch;
+}
+int MProcess::detectEFI(bool noTest)
+{
+    if (testEFI < 0) {
+        testEFI = 0;
+        QFile file("/sys/firmware/efi/fw_platform_size");
+        if (file.open(QFile::ReadOnly | QFile::Text)) {
+            testEFI = file.readLine().toInt();
+            file.close();
+        }
+        qDebug() << "Detect EFI:" << testEFI;
+    }
+    if (!noTest && testEFI == 64) {
+        // Bad combination: 32-bit OS on 64-bit EFI.
+        if (detectArch()=="i686") return 0;
+    }
+    return testEFI;
+}
+bool MProcess::detectMac()
+{
+    bool rc = false;
+    if (testMac < 0) {
+        // if it looks like an apple...
+        rc = exec("grub-probe", {"-d", "/dev/sda2"}, nullptr, true);
+        if (rc) rc = readOut(true).contains("hfsplus");
+        testMac = (rc ? 1 : 0);
+        qDebug() << "Detect Mac:" << rc;
+    }
+    return rc;
+}
